/****************************
 * Supporting math routines.
 ***************************/
 
#ifndef	__GNUC__ 
 
#include "stdafx.h" 
 
#endif 
 
#include <math.h>
#include "Supmath.h" 
#include "Newpgen2.h" 
#include "Primetable.h" 
 

 
#ifdef	__GNUC__ 
 
unsigned int _control87(unsigned int n, unsigned int mask) { 
	newfpcntrl = n & mask; 
	getcw87(); 
	newfpcntrl |= (oldfpcntrl & ~mask); 
	setcw87(); 
	return newfpcntrl; 
} 
 
#define TWOTO48M1 0xFFFFFFFFFFFFLL 
 
#else 
 
#include <float.h> 
 
#define TWOTO48M1 0xFFFFFFFFFFFF 
 
#endif 
 
__int64 MAXGUAR;

bool fFPURoundDown;

#define FPUROUNDDOWN {if (!fFPURoundDown) {fFPURoundDown = true; _control87 (_RC_DOWN, _MCW_RC); }}
#define FPUROUNDNEAR {if (fFPURoundDown) {fFPURoundDown = false; _control87 (_RC_NEAR, _MCW_RC); }}

// find_k064 - find k064 assuming
//   firstbit - top bit on n
//   nvalue - n value
//   k064, k164, k264, and k364 are global output variables
// this handles 6 independent k064 and p values in parallel for speed.
// This code is dependent upon cmovxx instructions in Athlon or Pentium III or higher
// For older CPUs, a different branching logic is required.
// To test, execute CPUID with eax set to 1.  If edx bit 15 is set, cmovxx is supported.
//
void find_k0()
{
	FPUROUNDDOWN;
	find_k0asm ();
}

void find_k0_bigp()
{
	FPUROUNDNEAR;
	find_k0_bigpasm ();
}

void find_k0_nocmov_bigp()
{
	FPUROUNDNEAR;
	find_k0_nocmov_bigpasm ();
}


// const unsigned int One = 1;

 /**********************
 * This calculates a*b (mod p), with a, b, p large (<63 bits)
 **********************/
__int64 mulmodp2 (const __int64 a, const __int64 b, const __int64 p)
{ 
//	static __int64 res; 
 
	a64 = a, b64 = b, p64 = p; 
 
	if (p <= 0x7FFFFFFF)
	{ 
		if (a64 >= p64) a64 %= p64;
		if (b64 >= p64) b64 %= p64;
		FPUROUNDDOWN;
		mulmodp2asm1(); 
	} 
	else 
	{ 
		FPUROUNDNEAR;
		mulmodp2asm2();
		if (res64 < 0) res64 += p;
	} 
	return res64; 
} 

 /**********************
 * This calculates a*b (mod p), with a, b, p large (<63 bits)
 **********************/
__int64 mulmodp2inv (const __int64 a, const __int64 b, const __int64 p, const double pinv)
{ 
//	static __int64 res; 
 
	a64 = a, b64 = b, p64 = p, pinverse = pinv; 
 
	if (p <= 0x7FFFFFFF)
	{ 
		if (a64 >= p64) a64 %= p64;
		if (b64 >= p64) b64 %= p64;
		FPUROUNDDOWN;
		mulmodp2invasm1(); 
	} 
	else 
	{ 
		FPUROUNDNEAR;
		mulmodp2invasm2();
		if (res64 < 0) res64 += p;
	} 
	return res64; 
} 


/**********************
 * This calculates x such that x = base^(-1) mod p
 **********************/
__int64 GetBaseInverse (const __int64 p)
{
	__int64 i = p-2;
	__int64 r=1, x=base;

#ifndef _M_IX86
	// Find base^(p-2) (mod p)

	for (i=p-2;i!=1;i>>=1)
	{
		if (i & 1)
			r = mulmodp (r, x, p);
		x = mulmodp (x, x, p);
	}
	r = mulmodp (r, x, p);

	// p may not really be prime... lets make sure that r really is base^(-1)....
	if (mulmodp (r, base, p) != 1) 
		return 0;

	return baseinverse = r;
#else

	p64 = p, i64 = i, r64 = r, x64 = x; 
 
	FPUROUNDNEAR;
	getbaseinverseasm(); 

	if (tmp64 != 1) 
		return baseinverse = 0;

	return baseinverse = r64;

#endif
}

/**********************
 * This calculates x such that x = n^(-1) mod p
 **********************/
__int64 Get48bitInverse (const __int64 n, const __int64 p)
{
	__int64 i = p-2;
	__int64 r=1, x=n;

#ifndef _M_IX86
	// Find n^(p-2) (mod p)

	for (i=p-2;i!=1;i>>=1)
	{
		if (i & 1)
			r = mulmodp (r, x, p);
		x = mulmodp (x, x, p);
	}
	r = mulmodp (r, x, p);

	// p may not really be prime... lets make sure that r really is base^(-1)....
	if (mulmodp (r, n, p) != 1) 
		return 0;

	return r;
#else

	if (p <= 0x7FFFFFFF)
	{
		p64 = p;
		n64 = n;

		FPUROUNDNEAR;
		get48bitinverseasm1 ();

		if (r32 < 0) r32 += (int)p;

		return r32;
	}
	else if (p < MAXGUAR)		// We can guarantee that p is prime up to here.
	{
		r64 = 0;
		p64 = p;
		n64 = n;

		FPUROUNDDOWN;
		get48bitinverseasm2 ();

		if (r64 < 0) r64 += p;

		return r64;
	}
	else
	{
		n64 = n, p64 = p, i64 = i, r64 = r, x64 = x; 
 
		FPUROUNDNEAR;
		get48bitinverseasm3 (); 
 
		if (tmp64 != 1) return 0; 
		else return r64; 
	} 
#endif
}


EXTERNC	unsigned int SmallPackPrimorial[20000];
EXTERNC	unsigned int nSmallPackPrimorial;

EXTERNC	unsigned __int64 PackPrimorial[10000];
EXTERNC	unsigned int nPackPrimorial;

extern double LogPrimorial;	// The natural logarithm of n# (or k#). 
 
EXTERNC	double convert; 

void FillMyPackPrimorial ()
{
	nPackPrimorial = 1;
	
	PackPrimorial[0] = 1;
	LogPrimorial = 0.0; 

	int i = nMyPrimorial;

	for (i--; i >= 0; i--)
	{
#ifndef __GNUC__
		unsigned __int64 res = 0x1FFFFFFFFFFFFFFF/MyPrimorial[i];
#else
		unsigned __int64 res = 0x1FFFFFFFFFFFFFFFLL/MyPrimorial[i];
#endif
		unsigned int j;

		LogPrimorial += log ((double)MyPrimorial[i]);

		for (j = 0; j < nPackPrimorial; j++)
		{
			if (res >= PackPrimorial[j])
			{
				PackPrimorial[j] *= MyPrimorial[i];
				goto next;
			}
		}
		PackPrimorial[nPackPrimorial++] = MyPrimorial[i];
		next:;
	}
}
 
void FillPackPrimorial (unsigned int n)
{
	nSmallPackPrimorial = nPackPrimorial = 1;
	
	PackPrimorial[0] = SmallPackPrimorial[0] = (mode == (MODE_LP|MODE_PRIMORIAL) || mode == (MODE_LM|MODE_PRIMORIAL)) ? 1 : 2;
	LogPrimorial =(mode == (MODE_LP|MODE_PRIMORIAL) || mode == (MODE_LM|MODE_PRIMORIAL)) ? 0.0 : log(2.0); 

	if (mode & MODE_PLUS5) primetable[1] = 1;
	if (mode & MODE_PLUS7) primetable[2] = 1;

	int i = 0;
	while (primetable[i] <= n) i++;

	for (i--; i >= 0; i--)
	{
		unsigned __int64 rem = 0, rem1 = 0, res;
		unsigned int r = 0x7FFFFFFF/primetable[i];
		unsigned int j;

		LogPrimorial += log ((double)primetable[i]);

		for (j = 0; j < nSmallPackPrimorial; j++)
		{
			if (r >= SmallPackPrimorial[j])
			{
				SmallPackPrimorial[j] *= primetable[i];
				goto next;
			}
		}
		SmallPackPrimorial[nSmallPackPrimorial++] = primetable[i];
		next:;

		for (j = 0; j < nPackPrimorial; j++)
		{
			res = PackPrimorial[j];
			res *= primetable[i];
			res += rem;

			if (primetable[i] > 0xFFFF && PackPrimorial[j] > 0xFFFFFFFF)
			{
				rem1 = PackPrimorial[j]>>16;
				rem1 *= primetable[i];
				rem1 >>= 48;
				rem1 <<= 16;
			}
			else
			{
				rem1 = 0;
			}
			
			PackPrimorial[j] = res & TWOTO48M1;
			rem = res >> 48;
			rem += rem1;
		}
		if (rem)
		{
			PackPrimorial[nPackPrimorial++] = rem;
		}
	}

	if (mode & MODE_PLUS5) primetable[1] = 5;
	if (mode & MODE_PLUS7) primetable[2] = 7;
}

void pair_GetPrimorialModP (const __int64 p1, const __int64 p2, __int64 &b1, __int64 &b2)
{
//	__int64 r1 = 1, r2 = 1;

	r164 = r264 = 1;
	p164 = p1, p264 = p2; 
 
	FPUROUNDDOWN;
	pair_getprimorialmodpasm(); 
 
	if (r164 == p164) r164 = 0; 
	if (r264 == p264) r264 = 0; 
 
	b1 = r164; 
	b2 = r264; 
}

__int64 GetPrimorialModP (const __int64 p1)
{
//	static unsigned __int64 tmp;
//	static const __int64 y = 0x1000000000000;
//	static float p_inverse;

	p164 = p1; 

	FPUROUNDDOWN;
	getprimorialmodpasm(); 

	if (tmp64>=p1) 
	{
		tmp64-=p1; 
	}

	return tmp64;
}	


/**********************
 * This calculates k.base^n (mod p), with p large
 **********************/
__int64 fastkbasepownmodp (__int64 k, unsigned int n, const __int64 p)
{
	unsigned int i;
	__int64 r=1, x=base;

	if (base % p == 0)
	{
		return 0;
	}
	if (p < n) n %= (p-1);

	if (n)
	{
		for (i=n;i!=1;i>>=1)
		{
			if (i & 1)
				r = mulmodp2 (r, x, p);
			x = mulmodp2 (x, x, p);
		}
		r = mulmodp2 (r, x, p);
		r = mulmodp2 (r, k, p);
		return r;
	}
	else
	{
		return k % p;
	}

}

/**********************
 * This calculates a^n (mod p), with p large
 **********************/
__int64 fastapowbmodp (__int64 a, __int64 b, const __int64 p)
{
	__int64 i, r=1, x=b;

	for (i=b; i!=1; i>>=1)
	{
		if (i & 1)
			r = mulmodp2 (r, x, p);
		x = mulmodp2 (x, x, p);
	}
	r = mulmodp2 (r, x, p);
	return r;
}


/**********************
 * Given n & p, find b such that k= b(mod p) ==> p|k.2^n-1;
 *                               k=-b(mod p) ==> p|k.2^n+1.
 **********************/
void pair_findbbase2(const __int64 p1, const __int64 p2, __int64 &r1, __int64 &r2)
{
//	static __int64 b1, b2; 
//	static unsigned int p1o2, p2o2; 
 
	p164 = p1, p264 = p2; 
 
	FPUROUNDDOWN;
	pair_findbbase2asm(); 
 
	r1 = b164; 
	r2 = b264; 
}

/**********************
 * Given n & p, find b such that k= b(mod p) ==> p|k.base^n-1;
 *                               k=-b(mod p) ==> p|k.base^n+1.
 **********************/
__int64 findbotherbases(const __int64 p)
{
	if (nvalue == 1) return baseinverse;

	__int64 b = baseinverse;

#ifndef _M_IX86

	unsigned int bit = firstbit;
	do
	{
		squaremodp (b, p);

		if (nvalue & bit)
		{
			b = mulmodp (b, baseinverse, p);
		}

		bit >>= 1;
	} while (bit);

	return b;
#else
//	static __int64 r; 
	p64 = p, b64 = b; 
 
	FPUROUNDNEAR;
	findbotherbasesasm();	 
 
	return r64; 
#endif
}

void pair_find2pownmodp (const __int64 p1, const __int64 p2, __int64 &r1, __int64 &r2)
{
//	__int64 b1 = 2, b2 = 2;
	p164 = p1, p264 = p2; 
	b164 = 2, b264 = 2; 

	FPUROUNDDOWN;
	pair_find2pownmodpasm(); 

	r1 = b164;
	r2 = b264;
	if (r1 < 0) r1 += p1;
	if (r2 < 0) r2 += p2;
	r1--;
	r2--;
	r1 = p1 - r1;
	r2 = p2 - r2;
	if (r1 & 1) r1 += p1;
	if (r2 & 1) r2 += p2;
	r1 >>= 1;
	r2 >>= 1;
}

unsigned int  findk2pownmodp (unsigned int k, unsigned int nval, __int64 p1)
{
//	__int64 b = 2;
	k3642 = k, nvalue32 = nval, p164 = p1; 
	b64 = 2; 

	FPUROUNDDOWN;
	findk2pownmodpasm(); 

	if (b64 < 0) b64 += p164;

	return (unsigned int)b64;
}


__int64 find2pownmodp (const __int64 p)
{
//	__int64 b = 2;
	p64 = p; 
	b64 = 2; 
//	static __int64 r;

	FPUROUNDNEAR;
	find2pownmodpasm(); 

	if (r64 > p)
	{
		r64 -= p;
	}

	r64--;
	r64 = p - r64;
	if (r64 & 1) r64 += p;
	r64 >>= 1;

	return r64;
}

__int64 findbasepownmodp (const __int64 p)
{
	p64 = p; 
	b64 = base; 

	FPUROUNDNEAR;
	findbasepownmodpasm(); 

	if (r64 > p)
	{
		r64 -= p;
	}

	r64--;
	if ((base&1) == 1)
	{
		r64--;
		if (r64 < 0) r64 += p;
	}

	r64 = p - r64;
	if (r64 & 1) r64 += p;
	r64 >>= 1;

	return r64;
}

__int64 halfpownmodp (int n, int topbit, const __int64 p)
{
	p64 = p; 
	b64 = (p+1)>>1; 

	firstbit2 = topbit;
	nvalue2 = n;

	FPUROUNDNEAR;

#ifndef __GNUC__

	_asm
	{
	fild	QWORD PTR p64
	fld1
	fdivrp	ST(1), ST(0)

	mov		eax, DWORD PTR b64
	mov		edx, DWORD PTR b64 + 4

	mov		DWORD PTR res64, eax
	mov		DWORD PTR res64+4, edx

	mov		ecx, DWORD PTR firstbit2
	mov		ebx, DWORD PTR nvalue2
loop0n:
	fild	QWORD PTR res64
	mul		eax
	fmul	ST(0), ST(0)
	mov		esi, eax
	mov		edi, edx
	mov		eax, DWORD PTR res64
	fmul	ST(0), ST(1)
	mul		DWORD PTR res64+4
	fistp	QWORD PTR res64
	add		edi, eax
	add		edi, eax
	mov		eax, DWORD PTR p64
	mul		DWORD PTR res64
	sub		esi, eax
	sbb		edi, edx
	mov		eax, DWORD PTR p64+4
	mul		DWORD PTR res64
	sub		edi, eax
	mov		eax, DWORD PTR p64
	mul		DWORD PTR res64+4
	sub		edi, eax
	jns		label10n
	add		esi, dword ptr p64
	mov		edx, dword ptr p64+4
	adc		edi, edx
label10n:
	mov		eax, esi

;	/**** edi:esi (and edi:eax) contains b^2 mod p *****/

	test	ecx, ebx
	je		label2n

	test	eax, 1
	jz		evenn
	add		eax, dword ptr p64
	adc		edi, dword ptr p64+4

evenn:
	shr		edi, 1
	rcr		eax, 1
	
label2n:
	mov		DWORD PTR res64, eax
	mov		DWORD PTR res64+4, edi
	shr		ecx, 1
	jne		loop0n

	test	edi, edi
	jge		label4n

	add		eax, DWORD PTR p64
	mov		edx, DWORD PTR p64+4
	adc		edi, edx
	mov		DWORD PTR res64, eax
	mov		DWORD PTR res64+4, edi

label4n:
	fstp	ST(0)
	}

#else

	findhalfpownmodpasm(); 

#endif

	return res64;
}
