// frmTestSplit.cpp : implementation file
//

#include "stdafx.h"
#include "newpgen.h"
#include "frmTestSplit.h"
#include "frmPCSpeed.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// frmTestSplit dialog


frmTestSplit::frmTestSplit(CWnd* pParent /*=NULL*/)
	: CDialog(frmTestSplit::IDD, pParent)
{
	//{{AFX_DATA_INIT(frmTestSplit)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}


void frmTestSplit::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(frmTestSplit)
	DDX_Control(pDX, IDC_EDIT2, m_nPCs);
	DDX_Control(pDX, IDC_EDIT1, m_File);
	DDX_Control(pDX, IDC_STATIC1, m_Static);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(frmTestSplit, CDialog)
	//{{AFX_MSG_MAP(frmTestSplit)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// frmTestSplit message handlers

BOOL frmTestSplit::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
	
	char *HelpText = 
"This will split a NewPGen output file up so that it can be tested on various machines. "
"Enter the name of the file to be split up, and the number of machines to spread the work among. "
"Then for each machine, enter its processor speed (MHz).\r\n"
"Each machine will be given a proportion of the file to test, relative to its speed - so faster machines get more work.";

	// TODO: Add your specialized code here and/or call the base class
	m_Static.SetWindowText (HelpText);

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void frmTestSplit::OnOK() 
{
	// TODO: Add extra validation here

	int i;
	char buff[1024];
	char ifilename[1024];

	m_File.GetWindowText (ifilename, 1024);

	if (!ifilename[0])
	{
		m_File.SetFocus ();
		return;
	}

	FILE *ifptr = fopen (ifilename, "r");
	if (!ifptr)
	{
		AfxMessageBox ("The input file could not be opened!");
		m_File.SetFocus ();
		return;
	}

	m_nPCs.GetWindowText (buff, 1024);
	if (atoi (buff) < 1 || atoi (buff) > 100) 
	{
		m_nPCs.SetWindowText ("");
		m_nPCs.SetFocus ();
		return;
	}

	int nFiles = atoi (buff);
	int *Speeds = new int[nFiles];

	for (i = 1; i <= nFiles;)
	{
		// Jean: frmPCSpeed just gets the speed. On exit: 
		// dlg.m_Button == IDCANCEL (Cancel or Back pressed) or IDOK.
		// dlg.m_SpeedMHz is the speed that was entered.
		// The dialog does not accept speeds of <33.

		// Also, it defaults to using the last speed that was entered.

		frmPCSpeed dlg (i, i == nFiles, (i == 1) ? 0 : Speeds[i-2]);

		dlg.DoModal ();

		if (dlg.m_Button == IDCANCEL)
		{
			i--;
			if (i == 0) return;
		}
		else
		{
			Speeds[i-1] = dlg.m_SpeedMHz;
			i++;
		}
	}

	if (AfxMessageBox ("The output files will be created - are you sure?", MB_ICONQUESTION|MB_YESNO|MB_DEFBUTTON2) == IDYES)
	{
		// Do the work!
		FILE **ofptrs = new FILE *[nFiles];
		int totspeed = 0;

		for (i = 0; i  <nFiles; i++)
		{
			char fname[1024];
			sprintf (fname, "%s_%d_%d", ifilename, Speeds[i], i+1);

			ofptrs[i] = fopen (fname, "w");

			if (!ofptrs[i])
			{
				AfxMessageBox ("One of the output files could not be opened!");
				while (i--)
				{
					fclose (ofptrs[i]);
				}
				fclose (ifptr);
				return;
			}
			totspeed += Speeds[i];
		}

		if (totspeed > RAND_MAX)
		{
			// Ah, we had better scale all values down proportionally...
			int sumspeed = 0;

			for (i = 0; i < nFiles; i++)
			{
				Speeds[i] = Speeds[i]*((double)RAND_MAX/totspeed);
			}
		}

		// OK, lets write out the first line to each file.

		this->BeginWaitCursor ();

		char buff[1024];

		fgets (buff, 1024, ifptr);
		for (i = 0; i < nFiles; i++) fputs (buff, ofptrs[i]);

		// Now do the rest of the lines.
		while (!feof (ifptr))
		{
			fgets (buff, 1024, ifptr);

			int random = rand () % totspeed;
			int sofar = 0;

			for (i = 0; i < nFiles; i++)
			{
				sofar += Speeds[i];
				if (sofar > random)
				{
					fputs (buff, ofptrs[i]);		/*# Should check for no disk space or error (EOF return)! */
					break;
				}
			}
		}

		for (i = 0; i < nFiles; i++)
		{
			fclose (ofptrs[i]);
		}
		fclose (ifptr);
		this->EndWaitCursor ();

		sprintf (buff, "All done. The output files have the speed appended to them, then a number - for example a %d MHz machine's file could be called '%s_%d_2'", Speeds[0], ifilename, Speeds[0]);
		AfxMessageBox (buff);

	}
	else
	{
		return;
	}

	CDialog::OnOK();
}
