#include "stdafx.h"
#include "supmath.h"

// global 12 K values for the SSE2 version, outputs from findk0xM_SSE2
__declspec(align(16)) __int64 k0k1[2];		// k0 and k1
__declspec(align(16)) __int64 k2k3[2];		// k2 and k3
__declspec(align(16)) __int64 k4k5[2];		// k4 and k5
__declspec(align(16)) __int64 k6k7[2];		// k6 and k7
__declspec(align(16)) __int64 k8k9[2];		// k8 and k9
__declspec(align(16)) __int64 k10k11[2];	// k10 and k11

// define 12 prime values used as inputs to findk0xM_SSE2 routine
__declspec(align(16)) __int64 p0p1[4];		// p0p1 (128 bits), p0p1 >> 32 (128 bits)
__declspec(align(16)) __int64 p2p3[4];		// p2p3, p2p3 >> 32
__declspec(align(16)) __int64 p4p5[4];		// p3p4, p3p4 >> 32
__declspec(align(16)) __int64 p6p7[4];		// p6p7, p6p7 >> 32
__declspec(align(16)) __int64 p8p9[4];		// p8p9, p8p9 >> 32
__declspec(align(16)) __int64 p10p11[4];	// p10p11, p10p11 >> 32

// define SSE2 one values, this is stored in xmm5 for the duration of findk0xM_SSE2
static  __declspec(align(16)) __int64 OneMask[2] = {0x1, 0x1};


#define Calc_P_Inverse_Macro(p_val, res_val1, res_val2)	\
	__asm	fild	QWORD PTR p_val				\
	__asm	fld1								\
	__asm	fdivrp	ST(1), ST(0)				\
	__asm	fstp	TBYTE ptr res_val1			\
	__asm	fild	QWORD PTR p_val+8			\
	__asm	fld1								\
	__asm	fdivrp	ST(1), ST(0)				\
	__asm	fstp	TBYTE ptr res_val2

	// compute SSE2 portion of  xmm0 = (a_val * a_val) mod p_val
	// p_val is assumed to have p followed by p >> 32
	// r_val is assumed to be recent store and require
	// qword access to load the values.
	// The result is returned in xmm0
	// do not alter xmm5
	// leave p_val in xmm7
#define SSE2_SqMod_QLoad(a_val, r_val, p_val)		\
	__asm	movdqa	xmm1, a_val					\
	__asm	movdqa	xmm7, p_val					\
	__asm	movdqa	xmm6, p_val+16				\
	__asm	movdqa	xmm0, xmm1					\
	__asm	psrlq	xmm1, 32					\
	__asm	pmuludq	xmm1, xmm0					\
	__asm	movq	xmm4, r_val					\
	__asm	movq	xmm3, r_val+8				\
	__asm	punpcklqdq	xmm4, xmm3				\
	__asm	movdqa	xmm2, xmm4					\
	__asm	movdqa	xmm3, xmm4					\
	__asm	psrlq	xmm4, 32					\
	__asm	pmuludq	xmm2, xmm6					\
	__asm	pmuludq	xmm4, xmm7					\
	__asm	pmuludq	xmm0, xmm0					\
	__asm	pmuludq	xmm3, xmm7					\
	__asm	paddd	xmm1, xmm1					\
	__asm	psubd	xmm1, xmm2					\
	__asm	psubd	xmm1, xmm4					\
	__asm	psubq	xmm0, xmm3					\
	__asm	psllq	xmm1, 32					\
	__asm	paddd	xmm0, xmm1					\
	__asm	pxor	xmm3, xmm3					\
	__asm	pcmpgtd	xmm3, xmm0					\
	__asm	pshufd	xmm3, xmm3, 0xf5			\
	__asm	pand	xmm3, xmm7					\
	__asm	paddq	xmm0, xmm3


	// compute SSE2 portion of  xmm0 = (a_val * a_val) mod p_val
	// p_val is assumed to have p followed by p >> 32
	// The result is returned in xmm0
	// do not alter xmm5
	// leave p_val in xmm7
#define SSE2_SqMod(a_val, r_val, p_val)		\
	__asm	movdqa	xmm7, p_val					\
	__asm	movdqa	xmm6, p_val+16				\
	__asm	movdqa	xmm1, a_val					\
	__asm	movdqa	xmm0, xmm1					\
	__asm	psrlq	xmm1, 32					\
	__asm	pmuludq	xmm1, xmm0					\
	__asm	movdqa	xmm4, r_val					\
	__asm	movdqa	xmm2, xmm4					\
	__asm	movdqa	xmm3, xmm4					\
	__asm	psrlq	xmm4, 32					\
	__asm	pmuludq	xmm2, xmm6					\
	__asm	pmuludq	xmm4, xmm7					\
	__asm	pmuludq	xmm0, xmm0					\
	__asm	pmuludq	xmm3, xmm7					\
	__asm	paddd	xmm1, xmm1					\
	__asm	psubd	xmm1, xmm2					\
	__asm	psubd	xmm1, xmm4					\
	__asm	psubq	xmm0, xmm3					\
	__asm	psllq	xmm1, 32					\
	__asm	paddd	xmm0, xmm1					\
	__asm	pxor	xmm3, xmm3					\
	__asm	pcmpgtd	xmm3, xmm0					\
	__asm	pshufd	xmm3, xmm3, 0xf5			\
	__asm	pand	xmm3, xmm7					\
	__asm	paddq	xmm0, xmm3

//
// Perform floating point part of mulmod
//  r_val = a_val*a_val*res_val1    and
//  r_val+8 = (a_val+8)*(a_val+8)*res_val2
// where res_val1 and res_val2 are 1/p for that calculation
// Do not touch any xmm registers.
//
#define FP_K0_Macro(a_val, r_val, res_val1, res_val2)	\
	__asm	fild	qword ptr a_val						\
	__asm	fmul	ST(0), ST(0)						\
	__asm	fld		TBYTE ptr res_val1					\
	__asm	fmulp	ST(1), ST(0)						\
	__asm	fistp	qword ptr r_val						\
	__asm	fild	qword ptr a_val+8					\
	__asm	fmul	ST(0), ST(0)						\
	__asm	fld		TBYTE ptr res_val2					\
	__asm	fmulp	ST(1), ST(0)						\
	__asm	fistp	qword ptr r_val+8

//
// perform odd test for the sieve
// if (k_val & 1)
//   k_val += r_val
// k_out = k_val >> 1
// same with kval2, r_val2, k_out2
// xmm5 is assumed to have two 64 bit values of 1, high and low
//

#define Odd_Test_Macro(k_val, r_val, k_val2, r_val2, k_out, k_out2)	\
	__asm	movdqa	xmm0, k_val							\
	__asm	movdqa	xmm1, r_val							\
	__asm	movdqa	xmm2, k_val2						\
	__asm	movdqa	xmm3, r_val2						\
	__asm	movdqa	xmm4, k_val							\
	__asm	pand	xmm0, xmm5							\
	__asm	pcmpeqd	xmm0, xmm5							\
	__asm	pshufd	xmm0, xmm0, 0xa0					\
	__asm	pand	xmm0, xmm1							\
	__asm	paddq	xmm4, xmm0							\
	__asm	psrlq	xmm4, 1								\
	__asm	movdqa	k_out, xmm4							\
	__asm	movdqa	xmm6, xmm2							\
	__asm	pand	xmm2, xmm5							\
	__asm	pcmpeqd	xmm2, xmm5							\
	__asm	pshufd	xmm2, xmm2, 0xa0					\
	__asm	pand	xmm2, xmm3							\
	__asm	paddq	xmm6, xmm2							\
	__asm	psrlq	xmm6, 1								\
	__asm	movdqa	k_out2, xmm6



//
// findk0xM_SSE2 - find k0 assuming
//   firstbit32 - top bit on n
//   nvalue - n value
//   p0p1, p2p3, p4p5, p6p7, p8p9, p10p11 are prime inputs
//   k0k1, k2k3, k4k5, k6k7, k8k9, k10k11 are output
// this handles 12  independent k0 and p values in parallel for speed.
// The 12x utilization allows for 128 bit access to variables since they are in
// cache and not just in the load/store unit which would require 64 bit access.
// This far overcomes the extra overhead of reloading the FP unit with 1/p
// (which happens on the underutilized port 0).
//

void __fastcall findk0xM_SSE2()
{
	static  __declspec(align(16)) __int64 r0r1[2];
	static  __declspec(align(16)) __int64 r2r3[2];
	static  __declspec(align(16)) __int64 r4r5[2];
	static  __declspec(align(16)) __int64 r6r7[2];
	static  __declspec(align(16)) __int64 r8r9[2];
	static  __declspec(align(16)) __int64 r10r11[2];

	// define temporary result values for FP part of the mulmod
	// these are stored as 10 byte, 80 bit floating point values
	static  __declspec(align(16)) __int64 Res1[2];
	static  __declspec(align(16)) __int64 Res2[2];
	static  __declspec(align(16)) __int64 Res3[2];
	static  __declspec(align(16)) __int64 Res4[2];
	static  __declspec(align(16)) __int64 Res5[2];
	static  __declspec(align(16)) __int64 Res6[2];
	static  __declspec(align(16)) __int64 Res7[2];
	static  __declspec(align(16)) __int64 Res8[2];
	static  __declspec(align(16)) __int64 Res9[2];
	static  __declspec(align(16)) __int64 Res10[2];
	static  __declspec(align(16)) __int64 Res11[2];
	static  __declspec(align(16)) __int64 Res12[2];

	// at the end of initialization
	// ecx will contain the bit count throughout the routine.
	// edx will contain nvalue
	// xmm5 will contain OneMask

	// prime the FP unit, calculate the inverse of the primes for later usage
	Calc_P_Inverse_Macro(p0p1, Res1, Res2)
	Calc_P_Inverse_Macro(p2p3, Res3, Res4)
	Calc_P_Inverse_Macro(p4p5, Res5, Res6)

	__asm	movdqa	xmm5, OneMask
	
	// set the initial k values while the FP unit works in the inverse operations.
	// using different sets of xmm registers allows more parallelism in this routine
	// that is why it was not converted to a macro.
	// k0 = (1+p0) >> 1
	// k1 = (1+p1) >> 1
	// set k0k1
	__asm	movdqa		xmm4, p0p1
	__asm	paddq		xmm4, xmm5
	__asm	psrlq		xmm4, 1
	__asm	movdqa		k0k1, xmm4

	// k2 = (1+p2) >> 1
	// k3 = (1+p3) >> 1
	// set k2k3
	__asm	movdqa		xmm6, p2p3
	__asm	paddq		xmm6, xmm5
	__asm	psrlq		xmm6, 1
	__asm	movdqa		k2k3, xmm6

	// k4 = (1+p5) >> 1
	// k5 = (1+p6) >> 1
	// set k4k5
	__asm	movdqa		xmm7, p4p5
	__asm	paddq		xmm7, xmm5
	__asm	psrlq		xmm7, 1
	__asm	movdqa		k4k5, xmm7

	// set k6k7
	__asm	movdqa		xmm4, p6p7
	__asm	paddq		xmm4, xmm5
	__asm	psrlq		xmm4, 1
	__asm	movdqa		k6k7, xmm4

	// set k8k9
	__asm	movdqa		xmm6, p8p9
	__asm	paddq		xmm6, xmm5
	__asm	psrlq		xmm6, 1
	__asm	movdqa		k8k9, xmm6

	// set k10k11
	__asm	movdqa		xmm7, p10p11
	__asm	paddq		xmm7, xmm5
	__asm	psrlq		xmm7, 1
	__asm	movdqa		k10k11, xmm7

	// continue the FP inverse operations
	Calc_P_Inverse_Macro(p6p7, Res7, Res8)
	Calc_P_Inverse_Macro(p8p9, Res9, Res10)
	Calc_P_Inverse_Macro(p10p11, Res11, Res12)

	__asm	mov		ecx, dword ptr firstbit32
	__asm	mov		edx, dword ptr nvalue
	
	// Main internal processing loop for exponentiation
	
MainLp:;
	// prime FP pipeline
	FP_K0_Macro(k0k1, r0r1, Res1, Res2)
	FP_K0_Macro(k2k3, r2r3, Res3, Res4)
	FP_K0_Macro(k4k5, r4r5, Res5, Res6)

	// branch mispredict is 20 cycles, every effort less than 20 cycles to avoid random
	// branching is taken.  Matches load and stores sizes when distance is short.
	
	// r0r1 is now just out of the FP unit.  Store forward it to the next macro
	// using qword access.
	// k0k1 = (k0k1*k0k1) mod p0p1
	SSE2_SqMod_QLoad(k0k1, r0r1, p0p1)
	__asm	movdqa	k0k1, xmm0
	
	// run the FP operations in parallel with the SSE2 operations
	FP_K0_Macro(k6k7, r6r7, Res7, Res8)
	FP_K0_Macro(k8k9, r8r9, Res9, Res10)
	FP_K0_Macro(k10k11, r10r11, Res11, Res12)

	// From here out, all values will be safely in cache and
	// 128 bit movdqa access can be used without any load/store penalties.
	// k2k3 = (k2k3*k2k3) mod p2p3
	SSE2_SqMod(k2k3, r2r3, p2p3)
	__asm	movdqa	k2k3, xmm0

	// k3k4 = (k4k5*k4k5) mod p4p5
	SSE2_SqMod(k4k5, r4r5, p4p5)
	__asm	movdqa	k4k5, xmm0

	SSE2_SqMod(k6k7, r6r7, p6p7)
	__asm	movdqa	k6k7, xmm0

	SSE2_SqMod(k8k9, r8r9, p8p9)
	__asm	movdqa	k8k9, xmm0

	SSE2_SqMod(k10k11, r10r11, p10p11)
	__asm	movdqa	k10k11, xmm0
	
	// test to see of an extra step is required in this exponentiation
	__asm	test	edx, ecx
	__asm	jnz		DoOddTest
	__asm	jmp		EndKTest
	
	DoOddTest:;
	
	// xmm5 has OneMask

	// if (k0k1 & 1)
	//		k0k1 += p0p1
	// k0k1 >> 1
	// if (k2k3 & 1)
	//		k2k3 += p2p3
	// k2k3 >> 1
	// ...
	// two sets (4 k values) are done in a single macro to
	// allow better xmm register utilization and parallelism
	Odd_Test_Macro(k0k1, p0p1, k2k3, p2p3, k0k1, k2k3)
	Odd_Test_Macro(k4k5, p4p5, k6k7, p6p7, k4k5, k6k7)
	Odd_Test_Macro(k8k9, p8p9, k10k11, p10p11, k8k9, k10k11)
		
	EndKTest:;
	__asm	shr		ecx, 1
	__asm	jnz		MainLp
}

