#include "Supmath.h"

// global 12 K values for the SSE2 version, outputs from findk0xM_SSE2

 __int64 k0k1[2] __attribute__((aligned(16)));		// k0 and k1
 __int64 k2k3[2] __attribute__((aligned(16)));		// k2 and k3
 __int64 k4k5[2] __attribute__((aligned(16)));		// k4 and k5
 __int64 k6k7[2] __attribute__((aligned(16)));		// k6 and k7
 __int64 k8k9[2] __attribute__((aligned(16)));		// k8 and k9
 __int64 k10k11[2] __attribute__((aligned(16)));	// k10 and k11

// define 12 prime values used as inputs to findk0xM_SSE2 routine

 __int64 p0p1[4] __attribute__((aligned(16)));		// p0p1 (128 bits), p0p1 >> 32 (128 bits)
 __int64 p2p3[4] __attribute__((aligned(16)));		// p2p3, p2p3 >> 32
 __int64 p4p5[4] __attribute__((aligned(16)));		// p3p4, p3p4 >> 32
 __int64 p6p7[4] __attribute__((aligned(16)));		// p6p7, p6p7 >> 32
 __int64 p8p9[4] __attribute__((aligned(16)));		// p8p9, p8p9 >> 32
 __int64 p10p11[4] __attribute__((aligned(16)));	// p10p11, p10p11 >> 32

// define SSE2 one values, this is stored in xmm5 for the duration of findk0xM_SSE2

static __int64 OneMask[2] __attribute__((aligned(16))) = {0x1, 0x1};

#if __GNUC__ >= 3

// N.B. : "fdivrp" replaced by "fdivp" in the next macro, due to a bug
// in gcc version 3.2 inline assembler...

#define Calc_P_Inverse_Macro(p_val, res_val1, res_val2)\
	asm ("		fildq	 	"#p_val"		");\
	asm ("		fld1					");\
	asm ("		fdivp	%st(0), %st(1)		");\
	asm ("		fstpt	 	"#res_val1"		");\
	asm ("		fildq		"#p_val"+8		");\
	asm ("		fld1					");\
	asm ("		fdivp	%st(0), %st(1)		");\
	asm ("		fstpt		"#res_val2"		")

	// compute SSE2 portion of  xmm0 = (a_val * a_val) mod p_val
	// p_val is assumed to have p followed by p >> 32
	// r_val is assumed to be recent store and require
	// qword access to load the values.
	// The result is returned in xmm0
	// do not alter xmm5
	// leave p_val in xmm7

#define SSE2_SqMod_QLoad(a_val, r_val, p_val)\
	asm ("		movdqa	"#a_val", %xmm1	");\
	asm ("		movdqa	"#p_val", %xmm7	");\
	asm ("		movdqa	"#p_val"+16, %xmm6");\
	asm ("		movdqa	%xmm1, %xmm0	");\
	asm ("		psrlq		$32, %xmm1		");\
	asm ("		pmuludq	%xmm0, %xmm1	");\
	asm ("		movq		"#r_val", %xmm4	");\
	asm ("		movq		"#r_val"+8, %xmm3	");\
	asm ("		punpcklqdq	%xmm3, %xmm4	");\
	asm ("		movdqa	%xmm4, %xmm2	");\
	asm ("		movdqa	%xmm4, %xmm3	");\
	asm ("		psrlq		$32, %xmm4		");\
	asm ("		pmuludq	%xmm6, %xmm2	");\
	asm ("		pmuludq	%xmm7, %xmm4	");\
	asm ("		pmuludq	%xmm0, %xmm0	");\
	asm ("		pmuludq	%xmm7, %xmm3	");\
	asm ("		paddd		%xmm1, %xmm1	");\
	asm ("		psubd		%xmm2, %xmm1	");\
	asm ("		psubd		%xmm4, %xmm1	");\
	asm ("		psubq		%xmm3, %xmm0	");\
	asm ("		psllq		$32, %xmm1		");\
	asm ("		paddd		%xmm1, %xmm0	");\
	asm ("		movdqa	%xmm0, k0k1		");\
	asm ("		pxor		%xmm3, %xmm3	");\
	asm ("		pcmpgtd	%xmm0, %xmm3	");\
	asm ("		pshufd	$245, %xmm3, %xmm3");\
	asm ("		pand		%xmm7, %xmm3	");\
	asm ("		paddq		%xmm3, %xmm0	")

	// compute SSE2 portion of  xmm0 = (a_val * a_val) mod p_val
	// p_val is assumed to have p followed by p >> 32
	// The result is returned in xmm0
	// do not alter xmm5
	// leave p_val in xmm7

#define SSE2_SqMod(a_val, r_val, p_val)\
	asm ("		movdqa	"#p_val", %xmm7	");\
	asm ("		movdqa	"#p_val"+16, %xmm6");\
	asm ("		movdqa	"#a_val", %xmm1	");\
	asm ("		movdqa	%xmm1, %xmm0	");\
	asm ("		psrlq		$32, %xmm1		");\
	asm ("		pmuludq	%xmm0, %xmm1	");\
	asm ("		movdqa	"#r_val", %xmm4	");\
	asm ("		movdqa	%xmm4, %xmm2	");\
	asm ("		movdqa	%xmm4, %xmm3	");\
	asm ("		psrlq		$32, %xmm4		");\
	asm ("		pmuludq	%xmm6, %xmm2	");\
	asm ("		pmuludq	%xmm7, %xmm4	");\
	asm ("		pmuludq	%xmm0, %xmm0	");\
	asm ("		pmuludq	%xmm7, %xmm3	");\
	asm ("		paddd		%xmm1, %xmm1	");\
	asm ("		psubd		%xmm2, %xmm1	");\
	asm ("		psubd		%xmm4, %xmm1	");\
	asm ("		psubq		%xmm3, %xmm0	");\
	asm ("		psllq		$32, %xmm1		");\
	asm ("		paddd		%xmm1, %xmm0	");\
	asm ("		pxor		%xmm3, %xmm3	");\
	asm ("		pcmpgtd	%xmm0, %xmm3	");\
	asm ("		pshufd	$245, %xmm3, %xmm3");\
	asm ("		pand		%xmm7, %xmm3	");\
	asm ("		paddq		%xmm3, %xmm0	")

//
// Perform floating point part of mulmod
//  r_val = a_val*a_val*res_val1    and
//  r_val+8 = (a_val+8)*(a_val+8)*res_val2
// where res_val1 and res_val2 are 1/p for that calculation
// Do not touch any xmm registers.
//

#define FP_K0_Macro(a_val, r_val, res_val1, res_val2)\
	asm ("		fildq	"#a_val"		");\
	asm ("		fmul	%st(0), %st(0)	");\
	asm ("		fldt	"#res_val1"		");\
	asm ("		fmulp	%st(0), %st(1)	");\
	asm ("		fistpq	"#r_val"	");\
	asm ("		fildq	"#a_val"+8		");\
	asm ("		fmul	%st(0), %st(0)	");\
	asm ("		fldt	"#res_val2"		");\
	asm ("		fmulp	%st(0), %st(1)	");\
	asm ("		fistpq	"#r_val"+8	")

//
// perform odd test for the sieve
// if (k_val & 1)
//   k_val += r_val
// k_out = k_val >> 1
// same with kval2, r_val2, k_out2
// xmm5 is assumed to have two 64 bit values of 1, high and low
//

#define Odd_Test_Macro(k_val, r_val, k_val2, r_val2, k_out, k_out2)\
	asm ("		movdqa	"#k_val", %xmm0	");\
	asm ("		movdqa	"#r_val", %xmm1	");\
	asm ("		movdqa	"#k_val2", %xmm2	");\
	asm ("		movdqa	"#r_val2", %xmm3	");\
	asm ("		movdqa	"#k_val", %xmm4	");\
	asm ("		pand		%xmm5, %xmm0	");\
	asm ("		pcmpeqd	%xmm5, %xmm0	");\
	asm ("		pshufd	$160,	%xmm0, %xmm0");\
	asm ("		pand		%xmm1, %xmm0	");\
	asm ("		paddq		%xmm0, %xmm4	");\
	asm ("		psrlq		$1, %xmm4		");\
	asm ("		movdqa	%xmm4, "#k_out"	");\
	asm ("		movdqa	%xmm2, %xmm6	");\
	asm ("		pand		%xmm5, %xmm2	");\
	asm ("		pcmpeqd	%xmm5, %xmm2	");\
	asm ("		pshufd	$160, %xmm2, %xmm2");\
	asm ("		pand		%xmm3, %xmm2	");\
	asm ("		paddq		%xmm2, %xmm6	");\
	asm ("		psrlq		$1, %xmm6		");\
	asm ("		movdqa	%xmm6, "#k_out2"	")

#else
#include "sse2mm.h"
#endif
//
// findk0xM_SSE2 - find k0 assuming
//   firstbit32 - top bit on n
//   nvalue - n value
//   p0p1, p2p3, p4p5, p6p7, p8p9, p10p11 are prime inputs
//   k0k1, k2k3, k4k5, k6k7, k8k9, k10k11 are output
//   k0k1sg, k2k3sg, k4k5sg, k6k7sg, k8k9sg, k10k11sg are output
// this handles 12  independent k0 and p values in parallel for speed.
// The 12x utilization allows for 128 bit access to variables since they are in
// cache and not just in the load/store unit which would require 64 bit access.
// This far overcomes the extra overhead of reloading the FP unit with 1/p
// (which happens on the underutilized port 0).
//
	static  __int64 r0r1[2] __attribute__((aligned(16)));
	static  __int64 r2r3[2] __attribute__((aligned(16)));
	static  __int64 r4r5[2] __attribute__((aligned(16)));
	static  __int64 r6r7[2] __attribute__((aligned(16)));
	static  __int64 r8r9[2] __attribute__((aligned(16)));
	static  __int64 r10r11[2] __attribute__((aligned(16)));

	// define temporary result values for FP part of the mulmod
	// these are stored as 10 byte, 80 bit floating point values

	static  __int64 Res1[2] __attribute__((aligned(16)));
	static  __int64 Res2[2] __attribute__((aligned(16)));
	static  __int64 Res3[2] __attribute__((aligned(16)));
	static  __int64 Res4[2] __attribute__((aligned(16)));
	static  __int64 Res5[2] __attribute__((aligned(16)));
	static  __int64 Res6[2] __attribute__((aligned(16)));
	static  __int64 Res7[2] __attribute__((aligned(16)));
	static  __int64 Res8[2] __attribute__((aligned(16)));
	static  __int64 Res9[2] __attribute__((aligned(16)));
	static  __int64 Res10[2] __attribute__((aligned(16)));
	static  __int64 Res11[2] __attribute__((aligned(16)));
	static  __int64 Res12[2] __attribute__((aligned(16)));


void /*__fastcall*/ findk0xM_SSE2()
{
	// at the end of initialization
	// ecx will contain the bit count throughout the routine.
	// edx will contain nvalue
	// xmm5 will contain OneMask

	// prime the FP unit, calculate the inverse of the primes for later usage

	Calc_P_Inverse_Macro(p0p1, Res1, Res2);
	Calc_P_Inverse_Macro(p2p3, Res3, Res4);
	Calc_P_Inverse_Macro(p4p5, Res5, Res6);

	asm ("		movdqa	OneMask, %xmm5	");
	
	// set the initial k values while the FP unit works in the inverse operations.
	// using different sets of xmm registers allows more parallelism in this routine
	// that is why it was not converted to a macro.
	// k0 = (1+p0) >> 1
	// k1 = (1+p1) >> 1
	// set k0k1

	asm ("		movdqa	p0p1, %xmm4		");
	asm ("		paddq		%xmm5, %xmm4	");
	asm ("		psrlq		$1, %xmm4		");
	asm ("		movdqa	%xmm4, k0k1		");

	// k2 = (1+p2) >> 1
	// k3 = (1+p3) >> 1
	// set k2k3

	asm ("		movdqa	p2p3, %xmm6		");
	asm ("		paddq		%xmm5, %xmm6	");
	asm ("		psrlq		$1, %xmm6		");
	asm ("		movdqa	%xmm6, k2k3		"); 

	// k4 = (1+p5) >> 1
	// k5 = (1+p6) >> 1
	// set k4k5

	asm ("		movdqa	p4p5, %xmm7		");
	asm ("		paddq		%xmm5, %xmm7	");
	asm ("		psrlq		$1, %xmm7		");
	asm ("		movdqa	%xmm7, k4k5		");

	// set k6k7

	asm ("		movdqa	p6p7, %xmm4		");
	asm ("		paddq		%xmm5, %xmm4	");
	asm ("		psrlq		$1, %xmm4		");
	asm ("		movdqa	%xmm4, k6k7		");

	// set k8k9

	asm ("		movdqa	p8p9, %xmm6		");
	asm ("		paddq		%xmm5, %xmm6	");
	asm ("		psrlq		$1, %xmm6		");
	asm ("		movdqa	%xmm6, k8k9		");

	// set k10k11

	asm ("		movdqa	p10p11, %xmm7	");
	asm ("		paddq		%xmm5, %xmm7	");
	asm ("		psrlq		$1, %xmm7		");
	asm ("		movdqa	%xmm7, k10k11	");

	// continue the FP inverse operations

	Calc_P_Inverse_Macro(p6p7, Res7, Res8);
	Calc_P_Inverse_Macro(p8p9, Res9, Res10);
	Calc_P_Inverse_Macro(p10p11, Res11, Res12);

	asm ("		movl		_firstbit32, %ecx	");
	asm ("		movl		_nvalue, %edx	");
	
	// Main internal processing loop for exponentiation
	
asm ("MainLp:");
	// prime FP pipeline

	FP_K0_Macro(k0k1, r0r1, Res1, Res2);
	FP_K0_Macro(k2k3, r2r3, Res3, Res4);
	FP_K0_Macro(k4k5, r4r5, Res5, Res6);

	// branch mispredict is 20 cycles, every effort less than 20 cycles to avoid random
	// branching is taken.  Matches load and stores sizes when distance is short.
	
	// r0r1 is now just out of the FP unit.  Store forward it to the next macro
	// using qword access.
	// k0k1 = (k0k1*k0k1) mod p0p1

	SSE2_SqMod_QLoad(k0k1, r0r1, p0p1);

	asm ("		movdqa	%xmm0, k0k1		");
	
	// run the FP operations in parallel with the SSE2 operations

	FP_K0_Macro(k6k7, r6r7, Res7, Res8);
	FP_K0_Macro(k8k9, r8r9, Res9, Res10);
	FP_K0_Macro(k10k11, r10r11, Res11, Res12);

	// From here out, all values will be safely in cache and
	// 128 bit movdqa access can be used without any load/store penalties.
	// k2k3 = (k2k3*k2k3) mod p2p3

	SSE2_SqMod(k2k3, r2r3, p2p3);

	asm ("		movdqa	%xmm0, k2k3		");

	// k3k4 = (k4k5*k4k5) mod p4p5

	SSE2_SqMod(k4k5, r4r5, p4p5);

	asm ("		movdqa	%xmm0, k4k5		");

	SSE2_SqMod(k6k7, r6r7, p6p7);

	asm ("		movdqa	%xmm0, k6k7		");

	SSE2_SqMod(k8k9, r8r9, p8p9);

	asm ("		movdqa	%xmm0, k8k9		");

	SSE2_SqMod(k10k11, r10r11, p10p11);

	asm ("		movdqa	%xmm0, k10k11	");
	
	// test to see of an extra step is required in this exponentiation

	asm ("		test	%ecx, %edx	");
	asm ("		jnz	DoOddTest	");
	asm ("		jmp	EndKTest	");
	
asm ("DoOddTest:");
	
	// xmm5 has OneMask

	// if (k0k1 & 1)
	//		k0k1 += p0p1
	// k0k1 >> 1
	// if (k2k3 & 1)
	//		k2k3 += p2p3
	// k2k3 >> 1
	// ...
	// two sets (4 k values) are done in a single macro to
	// allow better xmm register utilization and parallelism

	Odd_Test_Macro(k0k1, p0p1, k2k3, p2p3, k0k1, k2k3);
	Odd_Test_Macro(k4k5, p4p5, k6k7, p6p7, k4k5, k6k7);
	Odd_Test_Macro(k8k9, p8p9, k10k11, p10p11, k8k9, k10k11);
		
asm ("EndKTest:");

	asm ("		shr	$1, %ecx	");
	asm ("		jnz	MainLp	");
}

